@@ -2,11 +2,13 @@ package stats
 
 import (
 	"fmt"
+	"os"
 	"strings"
 	"sync/atomic"
 	"time"
 )
 
+//nolint:recvcheck
 //go:generate enumer -type=Type -text -transform=snake -output=./stats_type.go
 type Type int
 
@@ -34,7 +36,7 @@ func (s *Stats) Elapsed() time.Duration {
 	return time.Since(s.start)
 }
 
-func (s *Stats) Print() {
+func (s *Stats) PrintToStderr() {
 	components := []string{
 		"traversed %d files",
 		"emitted %d files for processing",
@@ -42,7 +44,8 @@ func (s *Stats) Print() {
 		"",
 	}
 
-	fmt.Printf(
+	_, _ = fmt.Fprintf(
+		os.Stderr,
 		strings.Join(components, "\n"),
 		s.Value(Traversed),
 		s.Value(Matched),
